/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "dawn.h"

typedef enum {
  ASSET_JSON_DATA_TYPE_OBJECT,
  ASSET_JSON_DATA_TYPE_ARRAY,
  ASSET_JSON_DATA_TYPE_STRING,
  ASSET_JSON_DATA_TYPE_NUMBER,
  ASSET_JSON_DATA_TYPE_BOOLEAN,
  ASSET_JSON_DATA_TYPE_NULL
} assetjsondatatype_t;

typedef struct _assetjson_t assetjson_t;

typedef struct _assetjson_t {
  assetjsondatatype_t type;

  union {
    struct {
      char_t **keys;
      assetjson_t **values;
      size_t length;
    } object;

    struct {
      assetjson_t **value;
      size_t length;
    } array;

    double_t number;
    char_t *string;
    bool_t boolean;
  };
} assetjson_t;

/**
 * Parses a JSON string.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParse(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a sub-object.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseSub(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a NULL value.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsNull(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a boolean value.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsBoolean(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as an object.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsObject(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as an array.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsArray(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a string.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsString(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a number.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonParseAsNumber(const char_t *json, assetjson_t **out);

/**
 * Parses a JSON string as a string.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
size_t assetJsonReadString(const char_t *json, char_t **out);

/**
 * Parses a JSON string as a number.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
assetjson_t *assetJsonGetObjectValue(assetjson_t *object, const char_t *key);

/**
 * Parses a JSON string as a number.
 * 
 * @param json JSON string to parse.
 * @param out Pointer to store the parsed JSON object.
 * @return The number of characters parsed.
 */
void assetJsonDispose(assetjson_t *json);