/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "drawui.h"
#include "drawshape.h"
#include "assert/assert.h"
#include "game/time.h"
#include "display/draw/drawtext.h"
#include "ui/textbox.h"
#include "ui/testmenu.h"
#include "util/math.h"
#include "util/memory.h"

void drawUIBox(
  const uint16_t x,
  const uint16_t y,
  const uint16_t width,
  const uint16_t height,
  const uint8_t color,
  const bool_t fill
) {
  assertTrue(x < FRAME_WIDTH, "Box is too far right.");
  assertTrue(y < FRAME_HEIGHT, "Box is too far down.");
  assertTrue(width > 2, "Box is too narrow.");
  assertTrue(height > 2, "Box is too short.");
  assertTrue(x + width <= FRAME_WIDTH, "Box is too wide.");
  assertTrue(y + height <= FRAME_HEIGHT, "Box is too tall.");

  // Draw the top and bottom borders
  for(uint16_t i = 0; i < width; i++) {
    FRAME_BUFFER[y * FRAME_WIDTH + x + i] = '-';
    FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x + i] = '-';
    FRAME_COLOR[y * FRAME_WIDTH + x + i] = color;
    FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x + i] = color;
  }

  // Draw the left and right borders
  for(uint16_t i = 0; i < height; i++) {
    FRAME_BUFFER[(y + i) * FRAME_WIDTH + x] = '|';
    FRAME_BUFFER[(y + i) * FRAME_WIDTH + x + width - 1] = '|';
    FRAME_COLOR[(y + i) * FRAME_WIDTH + x] = color;
    FRAME_COLOR[(y + i) * FRAME_WIDTH + x + width - 1] = color;
  }

  // Draw the corners
  FRAME_BUFFER[y * FRAME_WIDTH + x] = '+';
  FRAME_BUFFER[y * FRAME_WIDTH + x + width - 1] = '+';
  FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x] = '+';
  FRAME_BUFFER[(y + height - 1) * FRAME_WIDTH + x + width - 1] = '+';
  FRAME_COLOR[y * FRAME_WIDTH + x] = color;
  FRAME_COLOR[y * FRAME_WIDTH + x + width - 1] = color;
  FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x] = color;
  FRAME_COLOR[(y + height - 1) * FRAME_WIDTH + x + width - 1] = color;

  // Fill the inside
  if(!fill) return;
  drawBox(x + 1, y + 1, width - 2, height - 2, ' ', COLOR_BLACK);
}

void drawUIMenu(
  menu_t *menu,
  const uint16_t x,
  const uint16_t y,
  const uint16_t width,
  const uint16_t height,
  const uint8_t cursorColor,
  const uint8_t textColor
) {
  assertTrue(menu, "Menu cannot be NULL.");
  assertTrue(menu->rows > 0, "Rows must be greater than 0.");
  assertTrue(menu->columns > 0, "Columns must be greater than 0.");
  assertTrue(width*2 > menu->columns, "Width must be greater than columns.");

  uint16_t colWidth;
  uint16_t startingCol;
  uint16_t startingRow;
  uint16_t rowCount;
  uint16_t colCount;
  char_t *str;
  size_t i;

  // Determine the width of each column. This must include space for the cursor
  // and the text.
  if(menu->columns == 1 || rowCount == 0) {
    colWidth = width;
    colCount = 1;
  } else {
    size_t longestString = 0;
    i = 0;
    while(true) {
      str = menu->strings[i++];
      if(str == NULL) break;

      size_t len = strlen(str);
      if(len > longestString) longestString = len;

      // If we overrun the size of the box, we can stop looking.
      if((len+1) > width) {
        longestString = width;
        break;
      }
    }

    // How many columns can we fit?
    uint16_t possibleColumnsAtThisWidth = width / (longestString + 1);

    if(possibleColumnsAtThisWidth < menu->columns) {
      // We need to take a subset of the columns
      colWidth = width / possibleColumnsAtThisWidth;
      colCount = possibleColumnsAtThisWidth;
    } else {
      // We can fit all columns in
      colWidth = width / menu->columns;
      colCount = menu->columns;
    }

    assertTrue(longestString != 0, "Longest string has length 0?");
  }

  // Determine count of rows to render
  if(rowCount == 1) {
    rowCount = 1;
  } else {
    rowCount = mathMin(rowCount, height);
  }

  // Where are we rendering from?
  startingCol = menu->renderOffsetX;

  if(menu->x >= startingCol + colCount) {
    startingCol = (menu->x + 1) - colCount;
  } else if(menu->x < startingCol) {
    startingCol = menu->x;
  }

  startingRow = menu->renderOffsetY;
  if(menu->y >= startingRow + rowCount) {
    startingRow = (menu->y + 1) - rowCount;
  } else if(menu->y < startingRow) {
    startingRow = menu->y;
  }

  // Update back to the menu
  menu->renderOffsetX = startingCol;
  menu->renderOffsetY = startingRow;

  for(uint16_t col = startingCol; col < startingCol+colCount; col++) {
    for(uint16_t row = startingRow; row < startingRow+rowCount; row++) {
      uint16_t index = (row * menu->columns) + col;
      if(menu->strings[index] == NULL) continue;

      drawText(
        menu->strings[index],
        colWidth,
        x + (colWidth * (col - startingCol)) + 1,
        y + (row - startingRow),
        textColor
      );
    }
  }

  // Draw the cursor
  i = (
    x + ((menu->x - startingCol) * colWidth)
  ) + (
    (y + (menu->y - startingRow)) * FRAME_WIDTH
  );
  FRAME_BUFFER[i] = '>';
  FRAME_COLOR[i] = cursorColor;
}

void drawUIMenuBox(
  menu_t *menu,
  const uint16_t x,
  const uint16_t y,
  const uint16_t width,
  const uint16_t height,
  const uint8_t cursorColor,
  const uint8_t textColor,
  const uint8_t boxColor
) {
  drawUIBox(x, y, width, height, boxColor, true);
  drawUIMenu(
    menu,
    x + 1, y + 1,
    width - 2, height - 2,
    cursorColor, textColor
  );
}

void drawUITextbox() {
  if(!textboxIsOpen()) return;

  // Border
  drawUIBox(
    0, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT,
    DRAW_UI_TEXTBOX_WIDTH, DRAW_UI_TEXTBOX_HEIGHT,
    COLOR_MAGENTA, true
  );

  // Title
  drawText(
    TEXTBOX.title,
    -1,
    2, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT,
    COLOR_WHITE
  );

  // Text
  drawText(
    TEXTBOX.text,
    TEXTBOX.textIndex,
    1, FRAME_HEIGHT - DRAW_UI_TEXTBOX_HEIGHT + 1,
    COLOR_WHITE
  );

  // Blinking cursor
  memorySet(&FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS], ' ', 3);
  memorySet(&FRAME_COLOR[DRAW_UI_TEXTBOX_CURSOR_POS], COLOR_WHITE, 3);
  if(TEXTBOX.textIndex < TEXTBOX.textLength) return;
  
  int32_t blink = (int32_t)(TIME.time * DRAW_UI_TEXTBOX_BLINKS_PER_SECOND) % 2;
  FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS + 1] = blink ? '>' : ' ';
  FRAME_BUFFER[DRAW_UI_TEXTBOX_CURSOR_POS + 2] = blink ? ' ' : '>';
}