/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "entity.h"
#include "entitydirection.h"
#include "rpg/world/map.h"
#include "assert/assert.h"
#include "game/time.h"
#include "ui/textbox.h"
#include "util/memory.h"

void entityInit(
  entity_t *entity,
  const entitytype_t type,
  map_t *map
) {
  assertNotNull(entity, "Entity cannot be NULL.");
  assertNotNull(map, "Map cannot be NULL.");

  memorySet(entity, 0, sizeof(entity_t));
  entity->type = type;
  entity->map = map;

  switch(type) {
    case ENTITY_TYPE_NULL:
      assertUnreachable("Should not be initializing a NULL entity.");
      break;

    case ENTITY_TYPE_NPC:
      npcInit(&entity->npc);
      break;

    case ENTITY_TYPE_SIGN:
      signInit(&entity->sign);
      break;

    case ENTITY_TYPE_DOOR:
      doorInit(&entity->door);
      break;

    case ENTITY_TYPE_PLAYER:
      playerInit(entity);
      break;
  }
}

void entityUpdate(entity_t *entity) {
  assertNotNull(entity, "Entity cannot be NULL.");

  switch(entity->state) {
    case ENTITY_STATE_IDLE:
      switch(entity->type) {
        case ENTITY_TYPE_NULL:
          assertUnreachable("Should not be updating a NULL entity.");
          break;

        case ENTITY_TYPE_PLAYER:
          playerUpdate(entity);
          break;
      }
      break;

    case ENTITY_STATE_WALKING:
      entity->walk.time -= TIME.delta;
      if(entity->walk.time > 0.0f) return;
      entity->state = ENTITY_STATE_IDLE;
      break;

    case ENTITY_STATE_TALKING:
      if(!textboxIsOpen()) entity->state = ENTITY_STATE_IDLE;
      break;

    default:
      assertUnreachable("Unknown entity state.");
  }
}

void entityWalk(entity_t *entity, const entitydirection_t dir) {
  assertNotNull(entity, "Entity cannot be NULL.");

  uint16_t newX = entity->x, newY = entity->y;
  entityDirectionOffsetAdd(dir, &newX, &newY);

  // Update direction, this happens regardless of if the player can move
  entity->direction = dir;

  // Walking OOB?
  if(newX >= entity->map->width || newY >= entity->map->height) {
    return;
  }

  // Get tile at position
  tile_t tileAtPosition = mapTileGetByPosition(
    entity->map,
    newX,
    newY,
    entity->layer
  );
  if(tileIsSolid(tileAtPosition)) return;

  // Get entity at position
  entity_t *entityAtPosition = mapEntityGetByPosition(
    entity->map,
    newX,
    newY
  );
  if(entityAtPosition != NULL) return;

  // Commit to move
  entity->x = newX;
  entity->y = newY;

  entity->state = ENTITY_STATE_WALKING;
  entity->walk.time = 0.1f;

  // Check for triggers
  trigger_t *trigger = mapTriggerGetByPosition(
    entity->map,
    entity->x,
    entity->y
  );
  if(trigger != NULL) triggerOnStep(trigger, entity);
}

void entityPositionSet(entity_t *entity, const uint16_t x, const uint16_t y) {
  assertNotNull(entity, "Entity cannot be NULL.");
  entity->x = x;
  entity->y = y;
}