/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "map.h"
#include "assert/assert.h"
#include "util/memory.h"

map_t MAP_MAIN;

void mapInit(
  map_t *map,
  const maplist_t list,
  const uint16_t width,
  const uint16_t height,
  const uint8_t layers
) {
  assertNotNull(map, "Map cannot be NULL.");
  assertTrue(
    width > 0,
    "Map width must be greater than 0."
  );
  assertTrue(
    width <= MAP_WIDTH_MAX,
    "Map width must be less than or equal to MAP_WIDTH_MAX."
  );
  assertTrue(
    height > 0,
    "Map height must be greater than 0."
  );
  assertTrue(
    height <= MAP_HEIGHT_MAX,
    "Map height must be less than or equal to MAP_HEIGHT_MAX."
  );
  assertTrue(
    layers > 0,
    "Map layers must be greater than 0."
  );
  assertTrue(
    layers <= MAP_LAYERS_MAX,
    "Map layers must be less than or equal to MAP_LAYERS_MAX."
  );

  memorySet(map, 0, sizeof(map_t));

  map->list = list;
  map->width = width;
  map->height = height;
  map->layers = layers;
}

void mapUpdate(map_t *map) {
  for(uint8_t i = 0; i < map->entityCount; i++) {
    entity_t *entity = &map->entities[i];
    entityUpdate(entity);
  }
}

entity_t * mapEntityAdd(map_t *map) {
  assertTrue(
    map->entityCount < MAP_ENTITIES_MAX,
    "Cannot add any more entities to map."
  );

  entity_t *entity = &map->entities[map->entityCount];
  map->entityCount++;
  return entity;
}

entity_t * mapEntityGetByPosition(
  map_t *map,
  const uint16_t x,
  const uint16_t y
) {
  if(x >= map->width || y >= map->height) return NULL;

  for(uint8_t i = 0; i < map->entityCount; i++) {
    entity_t *entity = &map->entities[i];
    if(entity->x == x && entity->y == y) {
      return entity;
    }
  }

  return NULL;
}

entity_t * mapEntityGetByType(
  map_t *map,
  const entitytype_t type
) {
  for(uint8_t i = 0; i < map->entityCount; i++) {
    entity_t *entity = &map->entities[i];
    if(entity->type == type) return entity;
  }

  return NULL;
}

tile_t mapTileGetByPosition(
  const map_t *map,
  const uint16_t x,
  const uint16_t y,
  const uint8_t layer
) {
  if(x >= map->width || y >= map->height) return TILE_NULL;
  if(layer >= map->layers) return TILE_NULL;

  return map->tiles[
    (layer * map->width * map->height) +
    (y * map->width) +
    x
  ];
}

void mapTilesSet(
  map_t *map,
  const tile_t tiles[],
  const uint8_t layer
) {
  assertNotNull(map, "Map cannot be NULL.");
  assertNotNull(tiles, "Tiles cannot be NULL.");
  assertTrue(
    layer < map->layers,
    "Layer must be less than map layers."
  );
  
  memcpy(
    map->tiles + (layer * map->width * map->height),
    tiles,
    map->width * map->height * sizeof(tile_t)
  );
}

trigger_t * mapTriggerAdd(map_t *map) {
  assertTrue(
    map->triggerCount < MAP_TRIGGERS_MAX,
    "Cannot add any more triggers to map."
  );

  trigger_t *trigger = &map->triggers[map->triggerCount];
  map->triggerCount++;
  return trigger;
}

trigger_t * mapTriggerGetByPosition(
  map_t *map,
  const uint16_t x,
  const uint16_t y
) {
  for(uint8_t i = 0; i < map->triggerCount; i++) {
    trigger_t *trigger = &map->triggers[i];
    if(
      x >= trigger->x &&
      x < trigger->x + trigger->width &&
      y >= trigger->y &&
      y < trigger->y + trigger->height
    ) return trigger;
  }

  return NULL;
}