/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "menu.h"
#include "assert/assert.h"
#include "input.h"
#include "game/time.h"
#include "util/memory.h"

void menuInit(
  menu_t *menu,
  const uint16_t columns,
  const uint16_t rows
) {
  assertNotNull(menu, "Menu cannot be NULL.");
  assertTrue(rows > 0, "Rows must be greater than 0.");
  assertTrue(columns > 0, "Columns must be greater than 0.");
  memorySet(menu, 0, sizeof(menu_t));
  menu->rows = rows;
  menu->columns = columns;
}

void menuUpdate(menu_t *menu) {
  if(inputWasPressed(INPUT_BIND_UP)) {
    menu->repeatHeld = 0.0f;
    return menuPositionMove(menu, MENU_DIRECTION_UP);
  } else if(inputWasPressed(INPUT_BIND_DOWN)) {
    menu->repeatHeld = 0.0f;
    return menuPositionMove(menu, MENU_DIRECTION_DOWN);
  } else if(inputWasPressed(INPUT_BIND_LEFT)) {
    menu->repeatHeld = 0.0f;
    return menuPositionMove(menu, MENU_DIRECTION_LEFT);
  } else if(inputWasPressed(INPUT_BIND_RIGHT)) {
    menu->repeatHeld = 0.0f;
    return menuPositionMove(menu, MENU_DIRECTION_RIGHT);
  } else if(inputWasPressed(INPUT_BIND_ACCEPT) && menu->selectCallback) {
    menu->selectCallback(
      menu,
      menu->x, menu->y,
      menu->strings[(menu->y * menu->columns) + menu->x]
    );
    return;
  }

  // Handle repeat
  if(inputIsDown(INPUT_BIND_DOWN)) {
    menu->repeatHeld += TIME.delta;
    if(menu->repeatHeld >= MENU_HELD_TIME_INITIAL) {
      menu->repeatHeld = (MENU_HELD_TIME_INITIAL - MENU_HELD_TIME_REPEAT);
      menuPositionMove(menu, MENU_DIRECTION_DOWN);
    }
  } else if(inputIsDown(INPUT_BIND_UP)) {
    menu->repeatHeld += TIME.delta;
    if(menu->repeatHeld >= MENU_HELD_TIME_INITIAL) {
      menu->repeatHeld = (MENU_HELD_TIME_INITIAL - MENU_HELD_TIME_REPEAT);
      menuPositionMove(menu, MENU_DIRECTION_UP);
    }
  } else if(inputIsDown(INPUT_BIND_LEFT)) {
    menu->repeatHeld += TIME.delta;
    if(menu->repeatHeld >= MENU_HELD_TIME_INITIAL) {
      menu->repeatHeld = (MENU_HELD_TIME_INITIAL - MENU_HELD_TIME_REPEAT);
      menuPositionMove(menu, MENU_DIRECTION_LEFT);
    }
  } else if(inputIsDown(INPUT_BIND_RIGHT)) {
    menu->repeatHeld += TIME.delta;
    if(menu->repeatHeld >= MENU_HELD_TIME_INITIAL) {
      menu->repeatHeld = (MENU_HELD_TIME_INITIAL - MENU_HELD_TIME_REPEAT);
      menuPositionMove(menu, MENU_DIRECTION_RIGHT);
    }
  } else {
    menu->repeatHeld = 0.0f;
  }
}

void menuPositionMove(menu_t *menu, const menudirection_t direction) {
  uint16_t newY, newX;
  newX = menu->x;
  newY = menu->y;
  switch(direction) {
    case MENU_DIRECTION_UP:
      if(newY == 0) {
        newY = menu->rows - 1;
      } else {
        newY--;
      }
      break;

    case MENU_DIRECTION_DOWN:
      if(newY == menu->rows - 1) {
        newY = 0;
      } else {
        newY++;
      }
      break;

    case MENU_DIRECTION_LEFT:
      if(newX == 0) {
        newX = menu->columns - 1;
      } else {
        newX--;
      }
      break;

    case MENU_DIRECTION_RIGHT:
      if(newX == menu->columns - 1) {
        newX = 0;
      } else {
        newX++;
      }
      break;
  }

  menuPositionSet(menu, newX, newY);
}

void menuPositionSet(menu_t *menu, const uint16_t x, const uint16_t y) {
  assertNotNull(menu, "Menu cannot be NULL.");
  menu->x = x % menu->columns;
  menu->y = y % menu->rows;

  // Is there a string here?
  uint16_t index = (menu->y * menu->columns) + menu->x;
  if(menu->strings[index] == NULL) {
    if(menu->y == 0) {
      // idk yet
      menu->x = 0;
    } else {
      menu->y--;
    }
  }
}

size_t menuGetLongestStringLength(const menu_t *menu) {
  assertNotNull(menu, "Menu cannot be NULL.");
  size_t longest = 0;
  for(uint16_t i = 0; i < MENU_ITEM_COUNT_MAX; i++) {
    if(menu->strings[i] == NULL) continue;

    size_t len = strlen(menu->strings[i]);
    if(len > longest) longest = len;
  }
  return longest;
}