/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "memory.h"
#include "assert/assert.h"

uint64_t MEMORY_ALLOCATED;

void * memoryAllocate(const size_t size) {
  assertTrue(size > 0, "Memory size must be greater than 0.");
  assertTrue(size != ((size_t)-1), "Memory size is probably too large.");

  void *ptr = malloc(size);
  if(ptr == NULL) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(EXIT_FAILURE);
  }
  MEMORY_ALLOCATED++;
  return ptr;
}

void memoryFree(void *ptr) {
  MEMORY_ALLOCATED--;
  free(ptr);
}

void * memoryReallocate(void *ptr, const size_t size) {
  assertTrue(size > 0, "Memory size must be greater than 0.");
  assertNotNull(ptr, "Memory pointer cannot be NULL.");

  ptr = realloc(ptr, size);
  if(ptr == NULL) {
    fprintf(stderr, "Failed to reallocate memory.\n");
    exit(EXIT_FAILURE);
  }
  return ptr;
}

void memorySet(void *ptr, const uint8_t value, const size_t size) {
  assertNotNull(ptr, "Memory pointer cannot be NULL.");
  assertTrue(size > 0, "Memory size must be greater than 0.");
  memset(ptr, value, size);
}

void memoryCopy(void *dest, const void *src, const size_t size) {
  assertNotNull(dest, "Memory destination cannot be NULL.");
  assertNotNull(src, "Memory source cannot be NULL.");
  assertTrue(size > 0, "Memory size must be greater than 0.");
  memcpy(dest, src, size);
}

void memoryInit() {
  MEMORY_ALLOCATED = 0;
  memoryTestZero();
}

void memoryTestZero() {
  assertTrue(MEMORY_ALLOCATED == 0, "Memory leak detected.");
}