/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "display/display.h"
#include "assert/assert.h"
#include "display/shaders/simpletexturedshader.h"
#include "display/backbuffer.h"
#include "display/font.h"
#include "display/mesh.h"
#include "display/frame.h"
#include "display/primitives/quad.h"
#include "game/time.h"

mesh_t FRAMEBUFFER_MESH;

void displayInit() {
  simpleTexturedShaderInit();
  fontInit();
  frameInit();

  meshInit(&FRAMEBUFFER_MESH);
  meshCreateBuffers(
    &FRAMEBUFFER_MESH,
    FRAME_WIDTH * FRAME_HEIGHT * PRIMITIVE_QUAD_VERTICE_COUNT,
    FRAME_WIDTH * FRAME_HEIGHT * PRIMITIVE_QUAD_INDICE_COUNT
  );
}

void displayUpdate() {
  frameUpdate();

  // Write the frame buffer to the mesh
  size_t quadIndex = 0;
  vec4_t uvs;
  vec4_t positions;
  uint8_t color;
  for(uint32_t x = 0; x < FRAME_WIDTH; x++) {
    for(uint32_t y = 0; y < FRAME_HEIGHT; y++) {
      char_t c = FRAME_BUFFER[y * FRAME_WIDTH + x];
      if(c == '\0' || c == ' ') continue;
      
      color = FRAME_COLOR[y * FRAME_WIDTH + x];
      fontCharGetUV(c, uvs);
      
      positions[0] = x * 8;
      positions[1] = y * 8;
      positions[2] = positions[0] + 8;
      positions[3] = positions[1] + 8;

      primitiveQuadBuffer(
        &FRAMEBUFFER_MESH,
        (vec2_t){positions[0], positions[1]},
        (vec2_t){positions[2], positions[3]},
        (vec2_t){uvs[0], uvs[1]},
        (vec2_t){uvs[2], uvs[3]},
        COLOR4F_COLOR_MAP[color],
        quadIndex * PRIMITIVE_QUAD_VERTICE_COUNT,
        quadIndex * PRIMITIVE_QUAD_INDICE_COUNT
      );
      quadIndex++;
    }
  }

  // Prepare the shader
  backBufferBind();
  backBufferClear();

  mat4_t view, projection, model;

  glm_mat4_identity(projection);
  glm_ortho(0.0f, (float_t)FRAME_PIXEL_WIDTH, (float_t)FRAME_PIXEL_HEIGHT, 0.0f, 0.0f, 1.0f, projection);
  simpleTexturedShaderSetProjection(projection);

  glm_mat4_identity(view);
  glm_lookat((vec3_t){0.0f, 0.0f, 1.0f}, (vec3_t){0.0f, 0.0f, 0.0f}, (vec3_t){0.0f, 1.0f, 0.0f}, view);
  simpleTexturedShaderSetView(view);

  // Draw the frame mesh
  glm_mat4_identity(model);
  simpleTexturedShaderSetModel(model);

  simpleTexturedShaderBind();
  simpleTexturedShaderSetTexture(&FONT.texture);
  simpleTexturedShaderSetColor(COLOR4F_WHITE);

  meshDraw(
    &FRAMEBUFFER_MESH,
    MESH_DRAW_MODE_TRIANGLES,
    0, quadIndex * PRIMITIVE_QUAD_INDICE_COUNT
  );
}

void displayDispose() {
  simpleTexturedShaderDispose();
  fontDispose();
  meshDispose(&FRAMEBUFFER_MESH);
}