/**
 * Copyright (c) 2023 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "mesh.h"
#include "assert/assert.h"
#include "assert/assertgl.h"

int32_t MESH_ACTIVE_COUNT = 0;

void meshInit(mesh_t *mesh) {
  assertNotNull(mesh, "meshInit: Mesh cannot be null.");

  mesh->vertexBuffer = -1;
  mesh->indexBuffer = -1;
  mesh->vertexArray = -1;
  mesh->verticeCount = -1;
  mesh->indiceCount = -1;

  MESH_ACTIVE_COUNT++;
}

void meshCreateBuffers(
  mesh_t *mesh,
  const int32_t verticeCount,
  const int32_t indiceCount
) {
  assertNotNull(mesh, "meshCreateBuffers: Mesh cannot be null.");
  assertTrue(
    verticeCount > 0,
    "meshCreateBuffers: Vertice count must be greater than 0."
  );
  assertTrue(
    indiceCount > 0,
    "meshCreateBuffers: Indice count must be greater than 0."
  );

  // Dispose old buffers (if present).
  meshDestroyBuffers(mesh);

  mesh->verticeCount = verticeCount;
  mesh->indiceCount = indiceCount;
  size_t sizePos = sizeof(vec3_t) * verticeCount;
  size_t sizeInds = sizeof(meshindice_t) * indiceCount;
  size_t sizeCoords = sizeof(vec2_t) * verticeCount;
  size_t sizeColors = sizeof(color4f_t) * verticeCount;

  // Generate vertex array, I don't think I need to do this tbh.
  glGenVertexArrays(1, &mesh->vertexArray);
  assertNoGLError();
  glBindVertexArray(mesh->vertexArray);
  assertNoGLError();

  // Create some buffers, one for the vertex data, one for the indices
  GLuint buffer[2];
  glGenBuffers(2, buffer);
  assertNoGLError();
  mesh->vertexBuffer = buffer[0];
  assertTrue(
    mesh->vertexBuffer >= 0,
    "meshCreateBuffers: Failed to create vertex buffer"
  );
  mesh->indexBuffer = buffer[1];
  assertTrue(
    mesh->indexBuffer >= 0,
    "meshCreateBuffers: Failed to create index buffer"
  );

  // Buffer an empty set of data then buffer each component
  glBindBuffer(GL_ARRAY_BUFFER, mesh->vertexBuffer);
  assertNoGLError();
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->indexBuffer);
  assertNoGLError();
  
  glBufferData(GL_ARRAY_BUFFER, sizePos+sizeCoords+sizeColors, 0, GL_DYNAMIC_DRAW);
  assertNoGLError();

  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeInds, 0, GL_DYNAMIC_DRAW);
  assertNoGLError();

  // Setup the attrib pointers
  size_t offset = 0;
  glVertexAttribPointer(
    0, sizeof(vec3_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(0);
  assertNoGLError();

  offset += sizePos;
  glVertexAttribPointer(
    1, sizeof(vec2_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(1);
  assertNoGLError();

  offset += sizeCoords;
  glVertexAttribPointer(
    2, sizeof(color4f_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(2);
  assertNoGLError();
}

void meshDestroyBuffers(mesh_t *mesh) {
  assertNotNull(mesh, "meshDestroyBuffers: Mesh cannot be null.");

  glBindBuffer(GL_ARRAY_BUFFER, 0);
  assertNoGLError();
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
  assertNoGLError();

  if(mesh->vertexBuffer != -1) {
    glDeleteBuffers(1, &mesh->vertexBuffer);
    assertNoGLError();
    mesh->vertexBuffer = -1;
    mesh->verticeCount = -1;
  }

  if(mesh->indexBuffer != -1) {
    glDeleteBuffers(1, &mesh->indexBuffer);
    assertNoGLError();
    mesh->indexBuffer = -1;
    mesh->indiceCount = -1;
  }

  if(mesh->vertexArray != -1) {
    glDeleteVertexArrays(1, &mesh->vertexArray);
    assertNoGLError();
    mesh->vertexArray = -1;
  }
}

void meshBufferPositions(
  const mesh_t *mesh,
  const vec3_t *positions,
  const size_t position,
  const size_t count
) {
  assertNotNull(mesh, "meshBufferPositions: Mesh cannot be null");
  assertNotNull(
    positions,
    "meshBufferPositions: Positions cannot be null"
  );
  assertTrue(
    position >= 0 && position < mesh->verticeCount,
    "meshBufferPositions: Position must be within range"
  );
  assertTrue(
    position+count <= mesh->verticeCount,
    "meshBufferPositions: Position + count must be within range"
  );
  assertTrue(
    count > 0,
    "meshBufferPositions: Count must be greater than zero"
  );

  size_t offsetPositions = sizeof(vec3_t) * position;

  glBindBuffer(GL_ARRAY_BUFFER, mesh->vertexBuffer);
  assertNoGLError();
  glBufferSubData(
    GL_ARRAY_BUFFER,
    offsetPositions,
    sizeof(vec3_t) * count,
    (void*)positions
  );
  assertNoGLError();
}

void meshBufferCoordinates(
  const mesh_t *mesh,
  const vec2_t *coordinates,
  const size_t position,
  const size_t count
) {
  assertNotNull(mesh, "meshBufferCoordinates: Mesh cannot be null");
  assertNotNull(
    coordinates,
    "meshBufferCoordinates: Coordinates cannot be null"
  );
  assertTrue(
    position >= 0 && position < mesh->verticeCount,
    "meshBufferCoordinates: Position must be within range"
  );
  assertTrue(
    position+count <= mesh->verticeCount,
    "meshBufferCoordinates: Position + count must be within range"
  );
  assertTrue(
    count > 0,
    "meshBufferCoordinates: Count must be greater than zero"
  );

  size_t offsetCoordinates = (
    (sizeof(vec3_t) * mesh->verticeCount) +
    (sizeof(vec2_t) * position)
  );

  glBindBuffer(GL_ARRAY_BUFFER, mesh->vertexBuffer);
  assertNoGLError();
  glBufferSubData(
    GL_ARRAY_BUFFER,
    offsetCoordinates,
    sizeof(vec2_t) * count,
    (void*)coordinates
  );
  assertNoGLError();
}

void meshBufferColors(
  const mesh_t *mesh,
  const color4f_t *colors,
  const size_t position,
  const size_t count
) {
  assertNotNull(mesh, "meshBufferColors: Mesh cannot be null");
  assertNotNull(
    colors,
    "meshBufferColors: Colors cannot be null"
  );
  assertTrue(
    position >= 0 && position < mesh->verticeCount,
    "meshBufferColors: Position must be within range"
  );
  assertTrue(
    position+count <= mesh->verticeCount,
    "meshBufferColors: Position + count must be within range"
  );
  assertTrue(
    count > 0,
    "meshBufferColors: Count must be greater than zero"
  );

  size_t offsetColors = (
    (sizeof(vec3_t) * mesh->verticeCount) +
    (sizeof(vec2_t) * mesh->verticeCount) +
    (sizeof(color4f_t) * position)
  );

  glBindBuffer(GL_ARRAY_BUFFER, mesh->vertexBuffer);
  assertNoGLError();
  glBufferSubData(
    GL_ARRAY_BUFFER,
    offsetColors,
    sizeof(color4f_t) * count,
    (void*)colors
  );
  assertNoGLError();
}

void meshBufferIndices(
  const mesh_t *mesh,
  const meshindice_t *indices,
  const size_t position,
  const size_t count
) {
  assertNotNull(
    mesh,
    "meshBufferIndices: Mesh cannot be null"
  );
  assertNotNull(
    indices,
    "meshBufferIndices: Indices cannot be null"
  );
  assertTrue(
    position >= 0 && position < mesh->indiceCount,
    "meshBufferIndices: Position must be within range"
  );
  assertTrue(
    position+count <= mesh->indiceCount,
    "meshBufferIndices: Position + count must be within range"
  );
  assertTrue(
    count > 0,
    "meshBufferIndices: Count must be greater than zero"
  );

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->indexBuffer);
  assertNoGLError();
  glBufferSubData(
    GL_ELEMENT_ARRAY_BUFFER,
    sizeof(meshindice_t) * position,
    sizeof(meshindice_t) * count,
    (void*)indices
  );
  assertNoGLError();
}

void meshDraw(
  const mesh_t *mesh,
  const meshdrawmode_t mode,
  const int32_t indiceStart,
  const int32_t indiceCount
) {
  assertNotNull(mesh, "meshDraw: Mesh cannot be null");

  // Don't draw empty meshes.
  if(indiceCount == 0 || mesh->vertexBuffer == -1 || mesh->indexBuffer == -1) {
    return;
  }

  assertTrue(
    indiceCount > 0 || indiceCount == -1,
    "meshDraw: Invalid indiceCount"
  );
  
  // Re-Bind the buffers
  assertNoGLError();
  glBindVertexArray(mesh->vertexArray);
  assertNoGLError();
  glBindBuffer(GL_ARRAY_BUFFER, mesh->vertexBuffer);
  assertNoGLError();
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->indexBuffer);
  assertNoGLError();

  // Re-Calculate the attrib pointers. Will need to revist this later.
  size_t offset = 0;
  glVertexAttribPointer(
    0, sizeof(vec3_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void*)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(0);
  assertNoGLError();

  offset += sizeof(vec3_t) * mesh->verticeCount;
  glVertexAttribPointer(
    1, sizeof(vec2_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void*)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(1);
  assertNoGLError();

  offset += sizeof(vec2_t) * mesh->verticeCount;
  glVertexAttribPointer(
    2, sizeof(color4f_t) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void*)offset
  );
  assertNoGLError();
  glEnableVertexAttribArray(2);
  assertNoGLError();

  // Render the elements.
  glDrawElements(
    mode,
    indiceCount == -1 ? (mesh->indiceCount - indiceStart) : indiceCount,
    GL_UNSIGNED_INT,
    (void*)(sizeof(meshindice_t) * indiceStart)
  );
  assertNoGLError();
}

void meshDispose(mesh_t *mesh) {
  meshDestroyBuffers(mesh);
  MESH_ACTIVE_COUNT--;
}